/*
 *
 * GStreamer
 * Copyright (C) 1999-2001 Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gst/gst.h>
#include <math.h>

/* include this header if you want to use dynamic parameters
#include <gst/control/control.h>
*/

#include "gstosd.h"
#include "gstosdparam.h"

#define Y_FROM_RGB(r,g,b) ((9798 * (r) + 19235 * (g) + 3736 * (b)) >> 15)
#define U_FROM_BY(b,y) 	  ((16122 * ((b) - (y))) >> 15) + 128;
#define V_FROM_RY(r,y) 	  ((25203 * ((r) - (y))) >> 15) + 128;

static GstElementDetails plugin_details = {
  "On Screen Display",
  "Filter/Video/Osd",
  "On Screen Display for Raw Video",
  "P2P-VCR, C-Lab, University of Paderborn"
};

/* Arguments */
enum {
  ARG_0,
  ARG_OSD_BUFFER
};

GST_PAD_TEMPLATE_FACTORY (gst_osd_src_factory,
  "gst_osd_src",
  GST_PAD_SRC,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW (
   "gst_osd_src",
   "video/x-raw-rgb",
     "bpp",        GST_PROPS_INT (32),
     "depth",      GST_PROPS_INT (32),
     "endianness", GST_PROPS_INT (G_BYTE_ORDER),
     "red_mask",   GST_PROPS_INT (0xff0000),
     "green_mask", GST_PROPS_INT (0xff00),
     "blue_mask",  GST_PROPS_INT (0xff),
     "width",      GST_PROPS_INT_RANGE(0, G_MAXINT),
     "height",     GST_PROPS_INT_RANGE(0, G_MAXINT) 
  ),
  GST_CAPS_NEW (
   "gst_osd_src",
   "video/x-raw-yuv",
     "format",     GST_PROPS_FOURCC(GST_STR_FOURCC("YV12")),
     "width",      GST_PROPS_INT_RANGE(0, G_MAXINT),
     "height",     GST_PROPS_INT_RANGE(0, G_MAXINT) 
  )
)

GST_PAD_TEMPLATE_FACTORY (gst_osd_sink_factory,
  "gst_osd_sink",
  GST_PAD_SINK,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW (
   "gst_osd_sink",
   "video/x-raw-rgb",
     "bpp",        GST_PROPS_INT (32),
     "depth",      GST_PROPS_INT (32),
     "endianness", GST_PROPS_INT (G_BYTE_ORDER),
     "red_mask",   GST_PROPS_INT (0xff0000),
     "green_mask", GST_PROPS_INT (0xff00),
     "blue_mask",  GST_PROPS_INT (0xff),
     "width",      GST_PROPS_INT_RANGE(0, G_MAXINT),
     "height",     GST_PROPS_INT_RANGE(0, G_MAXINT)
  ),
  GST_CAPS_NEW (
   "gst_osd_sink",
   "video/x-raw-yuv",
     "format",     GST_PROPS_FOURCC(GST_STR_FOURCC("YV12")),
     "width",      GST_PROPS_INT_RANGE(0, G_MAXINT),
     "height",     GST_PROPS_INT_RANGE(0, G_MAXINT)
  )
)

static void	               gst_osd_class_init	         (GstOsdClass *klass);
static void	               gst_osd_init	                 (GstOsd *filter);
GType                          gst_osd_get_type                  (void);
static void                    gst_osd_dispose                   (GObject *object);
static void                    gst_osd_finalize                  (GObject *object);
static void	               gst_osd_set_property              (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static void	               gst_osd_get_property              (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);
static void	               gst_osd_chain                     (GstPad *pad, GstData *in_data);
static void                    gst_osd_chain_rgb                 (GstOsd *filter, GstData *in_data);
static void                    gst_osd_chain_rgb_opaque_mmx      (guchar* video_in, guchar* osd_in, guchar* video_out,
							          guint first_block, guint middle_block, guint last_block);
static void                    gst_osd_chain_rgb_opaque          (guchar* video_in, guchar* osd_in, guchar* video_out,
							          guint first_block, guint middle_block, guint last_block);
static void                    gst_osd_chain_rgb_transparent_mmx (guchar* video_in, guchar* osd_in, guchar* video_out,
								  guint first_block, guint middle_block, guint last_block, guchar* osd_reg_save);
static void                    gst_osd_chain_rgb_transparent     (guchar* video_in, guchar* osd_in, guchar* video_out,
							          guint first_block, guint middle_block, guint last_block);
static void                    gst_osd_chain_yv12                (GstOsd *filter, GstData *in_data);
static void                    gst_osd_chain_yv12_opaque         (guchar* video_in, guchar* osd_in, guchar* video_out,
		 					          guint first_block, guint middle_block, guint last_block);
static void                    gst_osd_chain_yv12_transparent    (guchar* video_in, guchar* osd_in, guchar* video_out,
							          guint first_block, guint middle_block, guint last_block,
			  					  guchar threshold1, guchar threshold2, guint width, guint offset);
static void                    gst_osd_scale_image               (GstOsd *filter);
static void                    gst_osd_scale_image_x             (GstOsd *filter, guchar *src, gint src_width, guchar *dest, gint dest_width);
static void                    gst_osd_scale_image_y             (GstOsd *filter, guchar *src, gint src_height, guchar *dest, gint dest_height);
static GstColorSpaceYUVTables* gst_colorspace_init_yuv           (long depth, long red_mask, long green_mask, long blue_mask);
static void                    gst_colorspace_yuv_to_rgb32       (GstColorSpaceYUVTables *tables, unsigned char *lum, unsigned char *cb,
								  unsigned char *cr, unsigned char *out, int rows, int cols);
static int                     number_of_bits_set                (unsigned long a);
static int                     free_bits_at_bottom               (unsigned long a);
static void                    gst_colorspace_rgb32_to_yuv       (unsigned char *src, unsigned char *ydst, unsigned char *udst,
								  unsigned char *vdst, unsigned char *adst, guint width, guint height);
static GstPadLinkReturn        gst_osd_src_connect               (GstPad *pad, GstCaps *caps);
static GstPadLinkReturn        gst_osd_sink_connect              (GstPad *pad, GstCaps *caps);

static GstElementClass *parent_class = NULL;

/* initialize the plugin's class */
static void
gst_osd_class_init (GstOsdClass *klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass*) klass;
  gstelement_class = (GstElementClass*) klass;

  parent_class = g_type_class_ref (GST_TYPE_ELEMENT);

  gobject_class->set_property = gst_osd_set_property;
  gobject_class->get_property = gst_osd_get_property;
  gobject_class->dispose = gst_osd_dispose;
  gobject_class->finalize = gst_osd_finalize;

  g_object_class_install_property (gobject_class, ARG_OSD_BUFFER,
				   g_param_spec_pointer ("osdbuffer",
							 "osdbuffer",
							 "Pointer to the osd bitmap (in rgba format)",
							 G_PARAM_READWRITE)
				   );

  GST_DEBUG ("class initialized");
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
gst_osd_init (GstOsd *filter)
{
  filter->sinkpad = gst_pad_new_from_template (GST_PAD_TEMPLATE_GET (gst_osd_sink_factory), "sink");
  gst_pad_set_link_function (filter->sinkpad, gst_osd_sink_connect);
  gst_pad_set_chain_function (filter->sinkpad, gst_osd_chain);
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);

  filter->srcpad = gst_pad_new_from_template (GST_PAD_TEMPLATE_GET (gst_osd_src_factory), "src");
  gst_pad_set_link_function (filter->srcpad, gst_osd_src_connect);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);

  filter->osd_in_mutex                     = g_mutex_new();
  filter->paramScaled.rgb_buffer           = NULL;
  filter->paramScaled.yv12_buffer          = NULL;

  filter->paramUnscaled.rgb_buffer         = NULL;
  filter->paramUnscaled.yv12_buffer        = NULL;
  filter->paramUnscaled.width              = 0;
  filter->paramUnscaled.height             = 0;
  filter->paramUnscaled.first_row          = 0;
  filter->paramUnscaled.rows_count         = 0;
  filter->paramUnscaled.is_totally_opaque  = FALSE;

  filter->osd_is_scaled                    = FALSE;
  filter->osd_is_rescaling_necessary       = FALSE;
  filter->osd_is_mmx_supported             = (gst_cpu_get_flags() & GST_CPU_FLAG_MMX);

  filter->format                           = 0;
  filter->sinkcaps                         = NULL;
  filter->gst_colorspace_yuv_tables        = gst_colorspace_init_yuv(32, 0xff, 0x00ff, 0x0000ff);

  GST_DEBUG ("object initialized");
}

GType
gst_osd_get_type (void)
{
  static GType plugin_type = 0;

  if (!plugin_type)
  {
    static const GTypeInfo plugin_info =
    {
      sizeof (GstOsdClass),
      NULL,
      NULL,
      (GClassInitFunc) gst_osd_class_init,
      NULL,
      NULL,
      sizeof (GstOsd),
      0,
      (GInstanceInitFunc) gst_osd_init,
    };
    plugin_type = g_type_register_static (GST_TYPE_ELEMENT,
	                                  "GstOsd",
	                                  &plugin_info, 0);
  }
  return plugin_type;
}

/* chain function
 * this function does the actual processing
 */
static void
gst_osd_chain (GstPad *pad, GstData *in_data)
{
  GstOsd *filter;
  GstCaps *caps;
  GstBuffer *in_buf = GST_BUFFER(in_data);

  g_return_if_fail (GST_IS_PAD (pad));
  g_return_if_fail (in_buf != NULL);

  filter = GST_OSD (GST_OBJECT_PARENT (pad));
  g_return_if_fail (GST_IS_OSD (filter));

  GST_DEBUG ("gst_osd_chain entered");

  caps = gst_pad_get_caps(pad);
  gst_caps_get_fourcc_int (caps, "format", &filter->format);

  if (GST_IS_EVENT (in_data))
  {
    GstEvent *event = GST_EVENT(in_data);

    switch (GST_EVENT_TYPE (event)) {
      default:
	gst_pad_event_default (filter->sinkpad, event);
	break;
    }
    return;
  }

  switch (filter->format)
  {
    case GST_MAKE_FOURCC('R', 'G', 'B', ' '):
      if (filter->paramScaled.rgb_buffer)
      {
	gst_osd_chain_rgb(filter, in_data);
      }
      else
      {
	gst_pad_push (filter->srcpad, in_data);
      }
      break;
    case GST_MAKE_FOURCC('Y', 'V', '1', '2'):
      if (filter->paramScaled.yv12_buffer)
      {
	gst_osd_chain_yv12(filter, in_data);
      }
      else
      {
	gst_pad_push (filter->srcpad, in_data);
      }
      break;
    default:
      g_warning("Unexpected colorspace.");
      break;
  }
}

static void
gst_osd_chain_rgb (GstOsd *filter, GstData *in_data)
{
  guint buffer_size;
  guint first_block, middle_block, last_block;
  guchar *video_out, *video_in, *osd_in, *osd_reg_save;
  GstBuffer *in_buf = GST_BUFFER(in_data);
  GstBuffer *out_buf;

  GST_DEBUG ("processing rgb buffer");

  if (filter->osd_is_rescaling_necessary)
  {
    gst_osd_scale_image (filter);
    filter->osd_is_rescaling_necessary = FALSE;
  }

  out_buf = gst_buffer_new();
  GST_BUFFER_SIZE(out_buf)      = GST_BUFFER_SIZE(in_buf);
  GST_BUFFER_TIMESTAMP(out_buf) = GST_BUFFER_TIMESTAMP(in_buf);
  GST_BUFFER_DATA(out_buf)      = g_malloc(GST_BUFFER_SIZE(out_buf));

  buffer_size  = GST_BUFFER_SIZE(in_buf) >> 2;
  
  g_mutex_lock(filter->osd_in_mutex);
  video_in     = (guchar *) GST_BUFFER_DATA(in_buf);
  video_out    = (guchar *) GST_BUFFER_DATA(out_buf);
  osd_in       = (guchar *) filter->paramScaled.rgb_buffer;
  osd_reg_save = (guchar *) filter->osd_reg_save;

  first_block  = (filter->paramScaled.first_row * filter->paramScaled.width);
  middle_block = (filter->paramScaled.rows_count * filter->paramScaled.width);
  last_block   = (buffer_size - middle_block - first_block);

  if (filter->paramUnscaled.is_totally_opaque)
  {
    if (filter->osd_is_mmx_supported)
    {
      gst_osd_chain_rgb_opaque_mmx(video_in, osd_in, video_out, first_block, middle_block, last_block);
    }
    else
    {
      gst_osd_chain_rgb_opaque(video_in, osd_in, video_out, first_block, middle_block, last_block);
    }
  }
  else
  {
    if (filter->osd_is_mmx_supported)
    {
      gst_osd_chain_rgb_transparent_mmx(video_in, osd_in, video_out, first_block, middle_block, last_block, osd_reg_save);
    }
    else
    {
      gst_osd_chain_rgb_transparent(video_in, osd_in, video_out, first_block, middle_block, last_block);
    }
  }
  g_mutex_unlock(filter->osd_in_mutex);
  
  gst_buffer_unref(in_buf);
  gst_pad_push (filter->srcpad, GST_DATA(out_buf));
}

static void
gst_osd_chain_rgb_opaque_mmx (guchar *video_in, guchar *osd_in, guchar *video_out,
			      guint first_block, guint middle_block, guint last_block)
{
  GST_DEBUG ("is totally opaque (mmx support enabled)");

  __asm__ __volatile__ (
  "      pushf                                                                  \n"
  "      cld                                                                    \n"
  "      rep movsl                                                              \n"
  "      popf                                                                   \n"
  : "=S" (video_in),
    "=c" (first_block),
    "=D" (video_out)
  : "S" (video_in),
    "c" (first_block),
    "D" (video_out)
  : "memory"
  );

  __asm__ __volatile__ (
  "      orl %%ecx, %%ecx                                                       \n"
  "      jz END1                                                                \n"
  "L1:   movl (%%eax), %%edx                                                    \n"
  "      movl %%edx, (%%edi)                                                    \n"
  "      addl $4, %%eax                                                         \n"
  "      addl $4, %%esi                                                         \n"
  "      addl $4, %%edi                                                         \n"
  "      loop L1                                                                \n"
  "END1:                                                                        \n"
  : "=S" (video_in),
    "=a" (osd_in),
    "=c" (middle_block),
    "=D" (video_out)
  : "S" (video_in),
    "a" (osd_in),
    "c" (middle_block),
    "D" (video_out)
  : "edx", "memory"
  );

  __asm__ __volatile__ (
  "      pushf                                                                  \n"
  "      cld                                                                    \n"
  "      rep movsl                                                              \n"
  "      popf                                                                   \n"
  : "=S" (video_in),
    "=c" (last_block),
    "=D" (video_out)
  : "S" (video_in),
    "c" (last_block),
    "D" (video_out)
  : "memory"
  );
}

static void
gst_osd_chain_rgb_opaque (guchar *video_in, guchar *osd_in, guchar *video_out,
			  guint first_block, guint middle_block, guint last_block)
{
  guint i;
  guint *video_out_int, *video_in_int, *osd_in_int;

  GST_DEBUG ("is totally opaque (mmx support disabled - using C compatibility version)");

  video_out_int = (guint *) video_out;
  video_in_int  = (guint *) video_in;
  osd_in_int    = (guint *) osd_in;

  for(i = 0; i < first_block; i++)
  {
    *video_out_int++ = *video_in_int++;
  }

  for(i = 0; i < middle_block; i++)
  {
    *video_out_int++ = *osd_in_int++;
    video_in_int++;
  }

  for(i = 0; i < last_block; i++)
  {
    *video_out_int++ = *video_in_int++;
  }
}

static void
gst_osd_chain_rgb_transparent_mmx (guchar *video_in, guchar *osd_in, guchar *video_out,
				   guint first_block, guint middle_block, guint last_block,
				   guchar *osd_reg_save)
{
  GST_DEBUG ("transparency mode (mmx support enabled)");

  __asm__ __volatile__ (
  "      pushf                                                                  \n"
  "      cld                                                                    \n"
  "      rep movsl                                                              \n"
  "      popf                                                                   \n"
  : "=S" (video_in),
    "=c" (first_block),
    "=D" (video_out)
  : "S" (video_in),
    "c" (first_block),
    "D" (video_out)
  : "memory"
  );

  __asm__ __volatile__ (
  "      orl %%ecx, %%ecx                                                       \n"
  "      jz END2                                                                \n"
  "      fsave (%%edx)                                                          \n"
  "      pushl %%ebx                                                            \n"
  "      pushl %%edx                                                            \n"
  "      pxor %%mm0, %%mm0                                                      \n"
  "L2:   movl (%%eax), %%ebx                                                    \n"
  "      andl $0xff000000, %%ebx                                                \n"
  "      orl %%ebx, %%ebx          # Is the current pixel transparent?          \n"
  "      setz %%dl                 # If so, set %%dl to 1...                    \n"
  "      pushl %%ecx                                                            \n"
  "      movl (%%esi), %%ecx       # ... and copy the pixel value to ecx        \n"
  "      cmpl $0xff000000, %%ebx   # Is the current pixel opaque?               \n"
  "      setz %%dh                 # If so, set %%dh to 1...                    \n"
  "      cmovzl (%%eax), %%ecx     # ... and copy the pixel value to ecx        \n"
  "      movl %%ecx, (%%edi)                                                    \n"
  "      popl %%ecx                                                             \n"
  "      orw %%dx, %%dx            # Is %%dx equal to zero?                     \n"
  "      jnz CONT                  # If so, we have finished our job for this   \n"
  "                                # pixel                                      \n"
  "      movd (%%eax), %%mm1       # Process non-transparent and non-opaque     \n"
  "      movd (%%esi), %%mm2       # pixel                                      \n"
  "      punpcklbw %%mm0, %%mm1                                                 \n"
  "      punpcklbw %%mm0, %%mm2                                                 \n"
  "      movl %%ebx, %%edx                                                      \n"
  "      shrl $8, %%edx                                                         \n"
  "      orl %%edx, %%ebx                                                       \n"
  "      movl %%ebx, %%edx                                                      \n"
  "      shrl $16, %%edx                                                        \n"
  "      orl %%edx, %%ebx                                                       \n"
  "      movd %%ebx, %%mm3                                                      \n"
  "      punpcklbw %%mm0, %%mm3                                                 \n"
  "      not %%ebx                                                              \n"
  "      movd %%ebx, %%mm4                                                      \n"
  "      pmullw %%mm3, %%mm1                                                    \n"
  "      punpcklbw %%mm0, %%mm4                                                 \n"
  "      psrlw $8, %%mm1                                                        \n"
  "      pmullw %%mm4, %%mm2                                                    \n"
  "      packuswb %%mm0, %%mm1                                                  \n"
  "      psrlw $8, %%mm2                                                        \n"
  "      packuswb %%mm0, %%mm2                                                  \n"
  "      paddb %%mm2, %%mm1                                                     \n"
  "      movd %%mm1, (%%edi)                                                    \n"
  "CONT: addl $4, %%eax                                                         \n"
  "      addl $4, %%esi                                                         \n"
  "      addl $4, %%edi                                                         \n"
  "loopl L2                                                                     \n"
  "      popl %%edx                                                             \n"
  "      popl %%ebx                                                             \n"
  "      frstor (%%edx)                                                         \n"
  "END2:                                                                        \n"
  : "=S" (video_in),
    "=a" (osd_in),
    "=c" (middle_block),
    "=d" (osd_reg_save),
    "=D" (video_out)
  : "S" (video_in),
    "a" (osd_in),
    "c" (middle_block),
    "d" (osd_reg_save),
    "D" (video_out)
  : "mm0", "mm1", "mm2", "mm3", "mm4", "mm5", "cc", "memory"
  );

  __asm__ __volatile__ (
  "      pushf                                                                  \n"
  "      cld                                                                    \n"
  "      rep movsl                                                              \n"
  "      popf                                                                   \n"
  : "=S" (video_in),
    "=c" (last_block),
    "=D" (video_out)
  : "S" (video_in),
    "c" (last_block),
    "D" (video_out)
  : "memory"
  );
}

static void
gst_osd_chain_rgb_transparent (guchar *video_in, guchar *osd_in, guchar *video_out,
			       guint first_block, guint middle_block, guint last_block)
{
  guint i;
  guint *video_out_int, *video_in_int, *osd_in_int;

  GST_DEBUG ("transparency mode (mmx support disabled - using C compatibility version)");

  video_out_int = (guint *) video_out;
  video_in_int  = (guint *) video_in;
  osd_in_int    = (guint *) osd_in;

  video_out_int = (guint *) video_out;
  for(i = 0; i < first_block; i++)
  {
    *video_out_int++ = *video_in_int++;
  }

  video_out = (guchar *) video_out_int;
  for(i = 0; i < middle_block; i++)
  {
    /* Blue */
    *video_out++ = (*(osd_in + 3) * *(osd_in + 0) + (255 - *(osd_in + 3)) * *video_in++) >> 8;
    
    /* Green */
    *video_out++ = (*(osd_in + 3) * *(osd_in + 1) + (255 - *(osd_in + 3)) * *video_in++) >> 8;
    
    /* Red */
    *video_out++ = (*(osd_in + 3) * *(osd_in + 2) + (255 - *(osd_in + 3)) * *video_in++) >> 8;

    *video_out++ = 0;
    video_in++;
    osd_in += 4;
  }

  video_out_int = (guint *) video_out;
  for(i = 0; i < last_block; i++)
  {
    *video_out_int++ = *video_in_int++;
  }
}

static void
gst_osd_chain_yv12 (GstOsd *filter, GstData *in_data)
{
  guint buffer_size;
  guint first_block, middle_block, last_block;
  guchar *video_out, *video_in, *osd_in;
  GstBuffer *out_buf;
  GstBuffer *in_buf = GST_BUFFER(in_data);

  GST_DEBUG ("processing yv12 buffer");

  if (filter->osd_is_rescaling_necessary)
  {
    gst_osd_scale_image (filter);
    filter->osd_is_rescaling_necessary = FALSE;
  }

  out_buf = gst_buffer_new();
  GST_BUFFER_SIZE(out_buf)      = GST_BUFFER_SIZE(in_buf);
  GST_BUFFER_TIMESTAMP(out_buf) = GST_BUFFER_TIMESTAMP(in_buf);
  GST_BUFFER_DATA(out_buf)      = g_malloc(GST_BUFFER_SIZE(out_buf));

  buffer_size = GST_BUFFER_SIZE(in_buf) / 1.5;

  g_mutex_lock(filter->osd_in_mutex);
  video_in     = (guchar *) GST_BUFFER_DATA(in_buf);
  video_out    = (guchar *) GST_BUFFER_DATA(out_buf);
  osd_in       = (guchar *) filter->paramScaled.yv12_buffer;

  first_block  = (filter->paramScaled.first_row * filter->paramScaled.width);
  middle_block = (filter->paramScaled.rows_count * filter->paramScaled.width);
  last_block   = (buffer_size - middle_block - first_block);

  if (filter->paramUnscaled.is_totally_opaque)
  {
    gst_osd_chain_yv12_opaque(video_in, osd_in, video_out, first_block, middle_block, last_block);
  }
  else
  {
    gst_osd_chain_yv12_transparent(video_in, osd_in, video_out, first_block, middle_block, last_block,
				   filter->paramUnscaled.yv12_transparency_threshold1,
				   filter->paramUnscaled.yv12_transparency_threshold2, filter->paramScaled.width,
				   GST_BUFFER_SIZE(in_buf));
  }
  g_mutex_unlock(filter->osd_in_mutex);
  
  gst_buffer_unref(in_buf);
  gst_pad_push (filter->srcpad, GST_DATA(out_buf));
}

static void
gst_osd_chain_yv12_opaque (guchar *video_in, guchar *osd_in, guchar *video_out,
			   guint first_block, guint middle_block, guint last_block)
{
  guint i;

  GST_DEBUG ("is totally opaque (mmx support disabled - using C compatibility version)");

  /* Copy the Y values first... */
  for(i = 0; i < first_block; i++)
  {
    *video_out++ = *video_in++;
  }

  for(i = 0; i < middle_block; i++)
  {
    *video_out++ = *osd_in++;
    video_in++;
  }

  for(i = 0; i < last_block; i++)
  {
    *video_out++ = *video_in++;
  }

  /* ... and then the U values... */
  first_block >>= 2;
  middle_block >>= 2;
  last_block >>= 2;

  for(i = 0; i < first_block; i++)
  {
    *video_out++ = *video_in++;
  }

  for(i = 0; i < middle_block; i++)
  {
    *video_out++ = *osd_in++;
    video_in++;
  }

  for(i = 0; i < last_block; i++)
  {
    *video_out++ = *video_in++;
  }

  /* ... and last but not least the V values */
  for(i = 0; i < first_block; i++)
  {
    *video_out++ = *video_in++;
  }

  for(i = 0; i < middle_block; i++)
  {
    *video_out++ = *osd_in++;
    video_in++;
  }

  for(i = 0; i < last_block; i++)
  {
    *video_out++ = *video_in++;
  }
}

static void
gst_osd_chain_yv12_transparent (guchar *video_in, guchar *osd_in, guchar *video_out,
				guint first_block, guint middle_block, guint last_block,
				guchar threshold1, guchar threshold2, guint width, guint offset)
{
  guint i;
  guchar *osd_in_transparency = osd_in + offset;
  guchar *osd_in_transparency_temp = osd_in_transparency;
  guint row = 0;

  GST_DEBUG ("transparency mode (mmx support disabled - using C compatibility version)");

  /* Copy the Y values first... */
  for(i = 0; i < first_block; i++)
  {
    *video_out++ = *video_in++;
  }

  for(i = 0; i < middle_block; i++)
  {
    if (*osd_in_transparency <= threshold1)
    {
      /* OSD is totally transparent */
      *video_out++ = *video_in++;
      osd_in++;
    }
    else if (*osd_in_transparency >= threshold2)
    {
      /* OSD is totally opaque */
      *video_out++ = *osd_in++;
      video_in++;
    }
    else
    {
      /* We're between the two thresholds */
      if ((i + row) % 2)
      {
	*video_out++ = *osd_in++;
	video_in++;
      }
      else
      {
	*video_out++ = *video_in++;
	osd_in++;
      }
    }

    row += !(i % width);
    osd_in_transparency++;
  }

  for(i = 0; i < last_block; i++)
  {
    *video_out++ = *video_in++;
  }

  /* ... and then the U values... */
  osd_in_transparency = osd_in_transparency_temp;
  first_block >>= 2;
  middle_block >>= 2;
  last_block >>= 2;

  for(i = 0; i < first_block; i++)
  {
    *video_out++ = *video_in++;
  }

  for(i = 0; i < middle_block; i++)
  {
    if (*osd_in_transparency <= threshold1)
    {
      *video_out++ = *video_in++;
      osd_in++;
    }
    else
    {
      *video_out++ = *osd_in++;
      video_in++;
    }

    osd_in_transparency += 2;
    osd_in_transparency += width * !((i << 1) % width);
  }

  for(i = 0; i < last_block; i++)
  {
    *video_out++ = *video_in++;
  }

  /* ... and last but not least the V values */
  osd_in_transparency = osd_in_transparency_temp;

  for(i = 0; i < first_block; i++)
  {
    *video_out++ = *video_in++;
  }

  for(i = 0; i < middle_block; i++)
  {
    if (*osd_in_transparency <= threshold1)
    {
      *video_out++ = *video_in++;
      osd_in++;
    }
    else
    {
      *video_out++ = *osd_in++;
      video_in++;
    }

    osd_in_transparency += 2;
    osd_in_transparency += width * !((i << 1) % width);
  }

  for(i = 0; i < last_block; i++)
  {
    *video_out++ = *video_in++;
  }
}

static void
gst_osd_scale_image (GstOsd *filter)
{
  guchar *buffer_src;
  guchar *buffer_intermediate;
  guchar *rgb_buffer;
  guchar *yv12_buffer;
  guint buffer_size;

  GST_DEBUG ("got a new osd image");

  g_mutex_lock(filter->osd_in_mutex);
  buffer_src = (guchar *) filter->paramUnscaled.rgb_buffer;

  filter->paramScaled.is_totally_opaque = filter->paramUnscaled.is_totally_opaque;
  filter->paramScaled.first_row  = (guint) ((gdouble) filter->paramScaled.height /
					    filter->paramUnscaled.height * filter->paramUnscaled.first_row);
  filter->paramScaled.rows_count = (guint) ((gdouble) filter->paramScaled.height /
					    filter->paramUnscaled.height * filter->paramUnscaled.rows_count);

  if (filter->paramUnscaled.rgb_buffer != NULL)
  {
    if ((filter->paramUnscaled.width != filter->paramScaled.width) ||
	(filter->paramUnscaled.height != filter->paramScaled.height))
    {
      GST_DEBUG ("scaling");

      /* Allocate Memory */
      if (filter->paramScaled.rgb_buffer)
      {
	g_free (filter->paramScaled.rgb_buffer);
	filter->paramScaled.rgb_buffer = NULL;
      }

      buffer_intermediate = g_malloc (filter->paramScaled.width * filter->paramUnscaled.rows_count * 4);
      rgb_buffer = g_malloc (filter->paramScaled.width * filter->paramScaled.rows_count * 4);

      gst_osd_scale_image_x(filter, buffer_src, filter->paramUnscaled.width, buffer_intermediate, filter->paramScaled.width);
      gst_osd_scale_image_y(filter, buffer_intermediate, filter->paramUnscaled.rows_count, rgb_buffer, filter->paramScaled.rows_count);

      g_free(buffer_intermediate);
      buffer_intermediate = NULL;

      if (filter->paramScaled.rgb_buffer)
      {
	g_free (filter->paramScaled.rgb_buffer);
      }

      filter->paramScaled.rgb_buffer = rgb_buffer;
    }
    else
    {
      GST_DEBUG ("scaling not necessary");

      if (filter->paramScaled.rgb_buffer)
      {
	g_free (filter->paramScaled.rgb_buffer);
      }

      filter->paramScaled.rgb_buffer = g_memdup(filter->paramUnscaled.rgb_buffer,
						filter->paramUnscaled.width * filter->paramUnscaled.rows_count * 4);
    }

    buffer_size = filter->paramScaled.width * filter->paramScaled.height;

    /* Colorspace conversion from RGB32 to YV12 */
    yv12_buffer = g_malloc(buffer_size * 2.5);

    gst_colorspace_rgb32_to_yuv(filter->paramScaled.rgb_buffer,
				yv12_buffer,
				yv12_buffer + buffer_size + (buffer_size >> 2),
				yv12_buffer + buffer_size,
				yv12_buffer + buffer_size + (buffer_size >> 1),
				filter->paramScaled.width, filter->paramScaled.height);

    if (filter->paramScaled.yv12_buffer)
    {
      g_free(filter->paramScaled.yv12_buffer);
    }

    filter->paramScaled.yv12_buffer = yv12_buffer;
  }
  else
  {
    if (filter->paramScaled.rgb_buffer)
    {
      g_free (filter->paramScaled.rgb_buffer);
    }

    if (filter->paramScaled.yv12_buffer)
    {
      g_free (filter->paramScaled.yv12_buffer);
    }

    filter->paramScaled.rgb_buffer = NULL;
    filter->paramScaled.yv12_buffer = NULL;
  }

  g_mutex_unlock(filter->osd_in_mutex);
}

static void
gst_osd_scale_image_x(GstOsd *filter, guchar *src, gint src_width, guchar *dest, gint dest_width)
{
  gint pixel_frac;
  gint src_pos_abs;
  gint dest_pos, dest_pos_abs;
  gint numerator;
  gint r, g, b, a;
  gint i;

  src_pos_abs = 0;
  dest_pos_abs = 0;

  for (i = 0; i < filter->paramUnscaled.rows_count; i++)
  {
    dest_pos = 0;
    pixel_frac = dest_width;

    while (dest_pos < dest_width)
    {
      r = g = b = a = 0;
      numerator = 0;

      while (numerator + pixel_frac <= src_width)
      {
	numerator += pixel_frac;
	b += pixel_frac * src[src_pos_abs++];
	g += pixel_frac * src[src_pos_abs++];
	r += pixel_frac * src[src_pos_abs++];
	a += pixel_frac * src[src_pos_abs++];
	pixel_frac = dest_width;
      }

      if (numerator < src_width)
      {
	b += (src_width - numerator) * src[src_pos_abs];
	g += (src_width - numerator) * src[src_pos_abs + 1];
	r += (src_width - numerator) * src[src_pos_abs + 2];
	a += (src_width - numerator) * src[src_pos_abs + 3];
	pixel_frac -= src_width - numerator;
      }

      dest_pos++;

      dest[dest_pos_abs++] = b / src_width;
      dest[dest_pos_abs++] = g / src_width;
      dest[dest_pos_abs++] = r / src_width;
      dest[dest_pos_abs++] = a / src_width;
    }
  }
}

static void
gst_osd_scale_image_y(GstOsd *filter, guchar *src, gint src_height, guchar *dest, gint dest_height)
{
  gint pixel_frac;
  gint src_pos_abs;
  gint dest_pos, dest_pos_abs;
  gint numerator;
  gint r, g, b, a;
  gint i;

  for (i = 0; i < filter->paramScaled.width; i++)
  {
    dest_pos = 0;
    src_pos_abs = i * 4;
    dest_pos_abs = i * 4;
    pixel_frac = dest_height;

    while (dest_pos < dest_height)
    {
      r = g = b = a = 0;
      numerator = 0;

      while (numerator + pixel_frac <= src_height)
      {
	numerator += pixel_frac;
	b += pixel_frac * src[src_pos_abs];
	g += pixel_frac * src[src_pos_abs + 1];
	r += pixel_frac * src[src_pos_abs + 2];
	a += pixel_frac * src[src_pos_abs + 3];
	pixel_frac = dest_height;
	src_pos_abs += filter->paramScaled.width * 4;
      }

      if (numerator < src_height)
      {
	b += (src_height - numerator) * src[src_pos_abs];
	g += (src_height - numerator) * src[src_pos_abs + 1];
	r += (src_height - numerator) * src[src_pos_abs + 2];
	a += (src_height - numerator) * src[src_pos_abs + 3];
	pixel_frac -= src_height - numerator;
      }

      dest[dest_pos_abs]     = b / src_height;
      dest[dest_pos_abs + 1] = g / src_height;
      dest[dest_pos_abs + 2] = r / src_height;
      dest[dest_pos_abs + 3] = a / src_height;

      dest_pos++;
      dest_pos_abs += filter->paramScaled.width * 4;
    }
  }
}

static GstColorSpaceYUVTables *
gst_colorspace_init_yuv(long depth, long red_mask, long green_mask, long blue_mask)
{
  int CR, CB, i;
  int *L_tab, *Cr_r_tab, *Cr_g_tab, *Cb_g_tab, *Cb_b_tab;
  long *r_2_pix_alloc;
  long *g_2_pix_alloc;
  long *b_2_pix_alloc;

  GstColorSpaceYUVTables *tables = g_malloc(sizeof(GstColorSpaceYUVTables));

  L_tab    = tables->L_tab = (int *)g_malloc(256*sizeof(int)); 
  Cr_r_tab = tables->Cr_r_tab = (int *)g_malloc(256*sizeof(int));
  Cr_g_tab = tables->Cr_g_tab = (int *)g_malloc(256*sizeof(int));
  Cb_g_tab = tables->Cb_g_tab = (int *)g_malloc(256*sizeof(int));
  Cb_b_tab = tables->Cb_b_tab = (int *)g_malloc(256*sizeof(int));

  r_2_pix_alloc = (long *)g_malloc(768*sizeof(long));
  g_2_pix_alloc = (long *)g_malloc(768*sizeof(long));
  b_2_pix_alloc = (long *)g_malloc(768*sizeof(long));

  if (L_tab == NULL ||
      Cr_r_tab == NULL ||
      Cr_g_tab == NULL ||
      Cb_g_tab == NULL ||
      Cb_b_tab == NULL ||
      r_2_pix_alloc == NULL ||
      g_2_pix_alloc == NULL ||
      b_2_pix_alloc == NULL)
  {
    fprintf(stderr, "Could not get enough memory in InitColorDither\n");
    exit(1);
  }

  for (i=0; i<256; i++)
  {
    L_tab[i] = i;
    CB = CR = i;
    CB -= 128; CR -= 128;
    Cr_r_tab[i] =  (0.419/0.299) * CR;
    Cr_g_tab[i] = -(0.299/0.419) * CR;
    Cb_g_tab[i] = -(0.114/0.331) * CB; 
    Cb_b_tab[i] =  (0.587/0.331) * CB;
  }

  /* 
   * Set up entries 0-255 in rgb-to-pixel value tables.
   */
  for (i = 0; i < 256; i++)
  {
    r_2_pix_alloc[i + 256] = i >> (8 - number_of_bits_set(red_mask));
    r_2_pix_alloc[i + 256] <<= free_bits_at_bottom(red_mask);
    g_2_pix_alloc[i + 256] = i >> (8 - number_of_bits_set(green_mask));
    g_2_pix_alloc[i + 256] <<= free_bits_at_bottom(green_mask);
    b_2_pix_alloc[i + 256] = i >> (8 - number_of_bits_set(blue_mask));
    b_2_pix_alloc[i + 256] <<= free_bits_at_bottom(blue_mask);
    /*
     * If we have 16-bit output depth, then we double the value
     * in the top word. This means that we can write out both
     * pixels in the pixel doubling mode with one op. It is 
     * harmless in the normal case as storing a 32-bit value
     * through a short pointer will lose the top bits anyway.
     * A similar optimisation for Alpha for 64 bit has been
     * prepared for, but is not yet implemented.
     */
    if(!(depth == 32) && !(depth == 24))
    {
      r_2_pix_alloc[i + 256] |= (r_2_pix_alloc[i + 256]) << 16;
      g_2_pix_alloc[i + 256] |= (g_2_pix_alloc[i + 256]) << 16;
      b_2_pix_alloc[i + 256] |= (b_2_pix_alloc[i + 256]) << 16;
    }
#ifdef SIXTYFOUR_BIT
    if(depth == 32)
    {
      r_2_pix_alloc[i + 256] |= (r_2_pix_alloc[i + 256]) << 32;
      g_2_pix_alloc[i + 256] |= (g_2_pix_alloc[i + 256]) << 32;
      b_2_pix_alloc[i + 256] |= (b_2_pix_alloc[i + 256]) << 32;
    }
#endif
  }

  /*
   * Spread out the values we have to the rest of the array so that
   * we do not need to check for overflow.
   */
  for (i = 0; i < 256; i++)
  {
    r_2_pix_alloc[i] = r_2_pix_alloc[256];
    r_2_pix_alloc[i+ 512] = r_2_pix_alloc[511];
    g_2_pix_alloc[i] = g_2_pix_alloc[256];
    g_2_pix_alloc[i+ 512] = g_2_pix_alloc[511];
    b_2_pix_alloc[i] = b_2_pix_alloc[256];
    b_2_pix_alloc[i+ 512] = b_2_pix_alloc[511];
  }

  tables->r_2_pix = r_2_pix_alloc + 256;
  tables->g_2_pix = g_2_pix_alloc + 256;
  tables->b_2_pix = b_2_pix_alloc + 256;

  return tables;
}

/*
 * How many 1 bits are there in the longword.
 * Low performance, do not call often.
 */
static int
number_of_bits_set(unsigned long a)
{
  if(!a) return 0;
  if(a & 1) return 1 + number_of_bits_set(a >> 1);
  return(number_of_bits_set(a >> 1));
}

/*
 * How many 0 bits are there at least significant end of longword.
 * Low performance, do not call often.
 */
static int
free_bits_at_bottom(unsigned long a)
{
  /* assume char is 8 bits */
  if(!a) return sizeof(unsigned long) * 8;
  if(((long)a) & 1l) return 0;
  return 1 + free_bits_at_bottom ( a >> 1);
}

static void
gst_colorspace_yuv_to_rgb32(GstColorSpaceYUVTables *tables, unsigned char *lum, unsigned char *cb,
			    unsigned char *cr, unsigned char *out, int rows, int cols)
{
  int L, CR, CB;
  unsigned int *row1, *row2;
  unsigned char *lum2;
  int x, y;
  int cr_r;
  int crb_g;
  int cb_b;
  int cols_2 = cols>>1;

  row1 = (guint32  *)out;
  row2 = row1 + cols;
  lum2 = lum + cols;
  for (y=rows>>1; y; y--)
  {
    for (x=cols_2; x; x--)
    {
      CR = *cr++;
      CB = *cb++;
      cr_r = tables->Cr_r_tab[CR];
      crb_g = tables->Cr_g_tab[CR] + tables->Cb_g_tab[CB];
      cb_b = tables->Cb_b_tab[CB];

      L = tables->L_tab[(int) *lum++];
      
      *row1++ = (tables->r_2_pix[L+cr_r] | tables->g_2_pix[L+crb_g] | tables->b_2_pix[L+cb_b]);

      L = tables->L_tab[(int) *lum++];

      *row1++ = (tables->r_2_pix[L+cr_r] | tables->g_2_pix[L+crb_g] | tables->b_2_pix[L+cb_b]);

      /*
       * Now, do second row.
       */

      L = tables->L_tab [(int) *lum2++];

      *row2++ = (tables->r_2_pix[L+cr_r] | tables->g_2_pix[L+crb_g] | tables->b_2_pix[L+cb_b]);

      L = tables->L_tab [(int) *lum2++];

      *row2++ = (tables->r_2_pix[L+cr_r] | tables->g_2_pix[L+crb_g] | tables->b_2_pix[L+cb_b]);
    }
    lum = lum2;
    row1 = row2;
    lum2 += cols;
    row2 += cols;
  }
}

static void
gst_colorspace_rgb32_to_yuv (unsigned char *src, 
		             unsigned char *ydst,
		             unsigned char *udst,
		             unsigned char *vdst,
			     unsigned char *adst,
			     guint width, guint height)
{
  int y;
  const int chrom_width = width >> 1;
  int Y;
  int b, g, r, a, v;

  for (y = height; y; y -= 2)
  {
    int i;

    for (i = chrom_width; i; i--)
    {
      b = *src++;
      g = *src++;
      r = *src++;
      *adst++ = *src++;

      Y = Y_FROM_RGB (r,g,b);

      *ydst++ = Y;
      *udst++ = U_FROM_BY (b,Y);
      v = V_FROM_RY (r, Y);
      *vdst++ = v > 255 ? 255 : v;

      b = *src++;
      g = *src++;
      r = *src++;
      *adst++ = *src++;

      *ydst++ = Y_FROM_RGB (r,g,b);
    }

    for (i = width; i; i--)
    {
      b = *src++;
      g = *src++;
      r = *src++;
      *adst++ = *src++;

      *ydst++ = Y_FROM_RGB (r,g,b);
    }
  }
}

/* this function handles the link with other plug-ins */
static GstPadLinkReturn
gst_osd_src_connect (GstPad *pad, GstCaps *caps)
{
  GstOsd *filter;
  GstCaps *peercaps = NULL;
  GstCaps *ourcaps, *to_intersect, *try_peercaps;
  guint32 format;
  
  filter = GST_OSD (gst_pad_get_parent (pad));

  /* we cannot operate if we didn't get src caps */
  ourcaps = filter->sinkcaps;

  if (!ourcaps)
  {
    return GST_PAD_LINK_DELAYED;
  }

  /* first see if we can do the format natively by filtering the peer caps 
   * with our incomming caps */
  peercaps = gst_caps_intersect (caps, ourcaps);
  if (peercaps)
  {
    /* see if the peer likes it too, it should as the caps say so.. */
    gst_caps_ref(peercaps);
    if (gst_pad_try_set_caps (filter->srcpad, peercaps) > 0)
    {
      /* gst_caps_unref (peercaps); */

      gst_caps_get_fourcc_int (peercaps, "format", &filter->format);
      if (filter->format == GST_STR_FOURCC ("RGB "))
      {
	guint bpp;
	gst_caps_get_int (peercaps, "bpp", &bpp);
      }

      return GST_PAD_LINK_OK;
    }
    /* gst_caps_unref (peercaps); */
  }

  return GST_PAD_LINK_REFUSED;
}

/* this function handles the link with other plug-ins */
static GstPadLinkReturn
gst_osd_sink_connect (GstPad *pad, GstCaps *caps)
{
  GstOsd *filter;
  GstPad *peer;
  GstCaps *peercaps = NULL;

  guint width;
  guint height;

  filter = GST_OSD (gst_pad_get_parent (pad));

  if (!GST_CAPS_IS_FIXED (caps))
  {
    return GST_PAD_LINK_DELAYED;
  }

  width = filter->paramScaled.width;
  height = filter->paramScaled.height;

  gst_caps_get_int (caps, "width", &filter->paramScaled.width);
  gst_caps_get_int (caps, "height", &filter->paramScaled.height);
  gst_caps_get_fourcc_int (caps, "format", &filter->format);

  filter->osd_is_rescaling_necessary = (width != filter->paramScaled.width) ||
    (height != filter->paramScaled.height);

  gst_caps_replace_sink (&filter->sinkcaps, caps);
  gst_pad_try_set_caps (filter->srcpad, caps);

  return GST_PAD_LINK_OK;
}

static void
gst_osd_set_property (GObject *object, guint prop_id,
		      const GValue *value, GParamSpec *pspec)
{
  GstOsd *filter;
  GstOsdParam *osd_param;

  guint *buffer;

  g_return_if_fail (GST_IS_OSD (object));
  filter = GST_OSD (object);

  switch (prop_id)
  {
    case ARG_OSD_BUFFER:
      osd_param = g_value_get_pointer(value);

      if (filter->paramUnscaled.rgb_buffer)
      {
	g_free (filter->paramUnscaled.rgb_buffer);
	filter->paramUnscaled.rgb_buffer = NULL;
      }

      filter->paramUnscaled.width                        = osd_param->width;
      filter->paramUnscaled.height                       = osd_param->height;
      filter->paramUnscaled.first_row                    = osd_param->first_row;
      filter->paramUnscaled.rows_count                   = osd_param->rows_count;
      
      if (osd_param->rgb_buffer)
      {
	filter->paramUnscaled.rgb_buffer = g_malloc(osd_param->width * osd_param->height * 4);
	memcpy(filter->paramUnscaled.rgb_buffer, osd_param->rgb_buffer, osd_param->width * osd_param->height * 4);
      }

      filter->paramUnscaled.yv12_buffer                  = NULL;
      filter->paramUnscaled.is_totally_opaque            = osd_param->is_totally_opaque;
      filter->paramScaled.is_totally_opaque              = osd_param->is_totally_opaque;
      filter->paramUnscaled.yv12_transparency_threshold1 = osd_param->yv12_transparency_threshold1;
      filter->paramUnscaled.yv12_transparency_threshold2 = osd_param->yv12_transparency_threshold2;

      filter->osd_is_scaled = FALSE;
      gst_osd_scale_image (filter);
      filter->osd_is_scaled = TRUE;
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_osd_get_property (GObject *object, guint prop_id,
		      GValue *value, GParamSpec *pspec)
{
  GstOsd *filter;

  g_return_if_fail (GST_IS_OSD (object));
  filter = GST_OSD (object);

  switch (prop_id)
  {
    case ARG_OSD_BUFFER:
      g_value_set_pointer(value, &filter->paramUnscaled);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_osd_dispose (GObject *object)
{
}

static void
gst_osd_finalize (GObject *object)
{
  GstOsd *filter;

  g_return_if_fail (GST_IS_OSD (object));
  filter = GST_OSD (object);

  if (filter->paramScaled.rgb_buffer)
  {
    g_free (filter->paramScaled.rgb_buffer);
    filter->paramScaled.rgb_buffer = NULL;
  }

  if (filter->paramScaled.yv12_buffer)
  {
    g_free (filter->paramScaled.yv12_buffer);
    filter->paramScaled.yv12_buffer = NULL;
  }

  if (filter->paramUnscaled.rgb_buffer)
  {
    g_free (filter->paramUnscaled.rgb_buffer);
    filter->paramUnscaled.rgb_buffer = NULL;
  }

  /* Free the memory used by the gst_colorspace_... routines */
  g_free(filter->gst_colorspace_yuv_tables->L_tab);
  filter->gst_colorspace_yuv_tables->L_tab = NULL;

  g_free(filter->gst_colorspace_yuv_tables->Cr_r_tab);
  filter->gst_colorspace_yuv_tables->Cr_r_tab = NULL;

  g_free(filter->gst_colorspace_yuv_tables->Cr_g_tab);
  filter->gst_colorspace_yuv_tables->Cr_g_tab = NULL;

  g_free(filter->gst_colorspace_yuv_tables->Cb_g_tab);
  filter->gst_colorspace_yuv_tables->Cb_g_tab = NULL;

  g_free(filter->gst_colorspace_yuv_tables->Cb_b_tab);
  filter->gst_colorspace_yuv_tables->Cb_b_tab = NULL;

  g_free(filter->gst_colorspace_yuv_tables->r_2_pix - 256);
  filter->gst_colorspace_yuv_tables->r_2_pix = NULL;

  g_free(filter->gst_colorspace_yuv_tables->g_2_pix - 256);
  filter->gst_colorspace_yuv_tables->g_2_pix = NULL;

  g_free(filter->gst_colorspace_yuv_tables->b_2_pix - 256);
  filter->gst_colorspace_yuv_tables->b_2_pix = NULL;

  g_mutex_free(filter->osd_in_mutex);
}

/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and pad templates
 * register the features
 */
static gboolean
plugin_init (GstPlugin *plugin)
{
  return gst_element_register(plugin, "osd", GST_RANK_NONE, GST_TYPE_OSD);
}

GST_PLUGIN_DEFINE (
  GST_VERSION_MAJOR,
  GST_VERSION_MINOR,
  "osd",
  "On Screen Display",
  plugin_init,
  VERSION,
  "LGPL",
  "Bla",
  "none",
  "University of Paderborn"
)
