#ifndef __GST_OSD_H__
#define __GST_OSD_H__

#include <gst/gst.h>
#include "gstosdparam.h"

G_BEGIN_DECLS

/* #define's don't like whitespacey bits */
#define GST_TYPE_OSD \
  (gst_osd_get_type())
#define GST_OSD(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_OSD,GstOsd))
#define GST_OSD_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_OSD,GstOsd))
#define GST_IS_OSD(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_OSD))
#define GST_IS_OSD_CLASS(obj) \
  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_OSD))

typedef struct _GstOsd       GstOsd;
typedef struct _GstOsdClass  GstOsdClass;
typedef enum   _GstOsdFormat GstOsdFormat;

/*
 *  The following struct is taken from the GStreamer colorspace plugin
 **/

typedef struct _GstColorSpaceYUVTables GstColorSpaceYUVTables;

struct _GstColorSpaceYUVTables {
  int gammaCorrectFlag;
  double gammaCorrect;
  int chromaCorrectFlag;
  double chromaCorrect;

  int *L_tab, *Cr_r_tab, *Cr_g_tab, *Cb_g_tab, *Cb_b_tab;

  /*
   *  We define tables that convert a color value between -256 and 512
   *  into the R, G and B parts of the pixel. The normal range is 0-255.
   **/

  long *r_2_pix;
  long *g_2_pix;
  long *b_2_pix;
};

struct _GstOsd
{
  GstElement element;

  GMutex *osd_in_mutex;

  GstPad *sinkpad, *srcpad;

  GstOsdParam paramScaled;
  GstOsdParam paramUnscaled;

  guint32 format;

  gboolean osd_is_rescaling_necessary;
  gboolean osd_is_scaled;
  gboolean osd_is_mmx_supported;

  guchar osd_reg_save[108];

  GstColorSpaceYUVTables *gst_colorspace_yuv_tables;

  GstCaps *sinkcaps;
};

struct _GstOsdClass 
{
  GstElementClass parent_class;
};

GType gst_osd_get_type (void);

G_END_DECLS

#endif /* __GST_OSD_H__ */
